// C Source File
// Created 26/06/03; 14:28:43

#include "globals.h"

const unsigned char mask[2] = {P1_MASK,P2_MASK};

unsigned short test_mask,tot;
unsigned short init_coords[4];

unsigned short (*test) (const short i,const short j);		// contains pointers to tests function in order to have only one main function that could be used in many ways
unsigned short test_contact (const short i,const short j) { return (((table[i][j])&COLOR_MASK)==color)&&!(table[i][j]&TEST_MASK);}
unsigned short test_fill (const short i,const short j) { return !((table[i][j])&(FILL_MASK|mask[player])); }
unsigned short test_color_contact (const short i,const short j) { return (table[i][j]&=DATA_MASK)==(player_color[player]|mask[player]);}
unsigned short test_color_fill (const short i,const short j) { return !((table[i][j])&FILL_MASK);}

static void do_frag(const short i,const short j,const short I,const short J)
{
/*  |  I,J are the coordinates of the parent tile of the grid, i,j the ones of the child.
    |  if the child hasn't been tested yet (TEST_MARK) there is a probabilty (defined by the
    |  global var 'fragmentation' that it will be colorized with it parent's color. Then it
    |  is marked as tested with the flag.
*/

	if(!((table[i][j]&TEST_MASK) || random(2*fragmentation+3)))
		table[i][j] = table[I][J]|TEST_MASK;
}

void fill_table(void)
{
/*  |  This function fills the table with random colors, and then calls do_frag  the fragmentation.
    |  The colors of the two players must be different, so it will fix them too.
    |  Then it marks the two first tiles for the players and call the expand function to
    |  capture the tiles in contact.
*/

short i,j;
init_coords[0] = init_coords[3] = 0;
init_coords[1] = TABLE_HEIGHT-1;
init_coords[2] = TABLE_WIDTH-1;

	for(j=TABLE_HEIGHT;j--;)
		for(i=TABLE_WIDTH-((j&1));i--;)
			table[i][j] = (unsigned char)random(COLOR_MASK);  // fill the grid with random colors

	for(j=TABLE_HEIGHT;j--;)    // test each tile
		for(i=TABLE_WIDTH-(j&1);i--;)
			if(i+(j&1)+1<TABLE_WIDTH && j+1<TABLE_HEIGHT && i>0 && j>0)  // the tile should be in the grid
			{
				do_frag(i+(j&1),j-1,i,j);    // do the fragmentation
				do_frag(i-1+(j&1),j-1,i,j);
				do_frag(i+(j&1),j+1,i,j);
				do_frag(i-1+(j&1),j+1,i,j);
			}

  // the two players cannot begin with the same color
  while(table[0][TABLE_HEIGHT-1] == table[TABLE_WIDTH-1][0]) table[TABLE_WIDTH-1][0]=(unsigned char)random(COLOR_MASK);

  player_color[PLAYER1] = table[0][TABLE_HEIGHT-1]&COLOR_MASK;    // save each player's color
  player_color[PLAYER2] = table[TABLE_WIDTH-1][0]&COLOR_MASK;

  table[0][TABLE_HEIGHT-1] |= P1_MASK;          // player 1 : first tile
  table[TABLE_WIDTH-1][0] |= P2_MASK;           // player 2

  color = player_color[PLAYER2];
  player = PLAYER2;
  expand();      // capture tiles in contact with that first tile

  color = player_color[PLAYER1];
  player = PLAYER1;
  expand();
}

void draw_table(short X, short Y)
{
/*  |  display to the screen the grid, with fronteers around each players zones and around the grid
    |  at coords (X,Y). Counts the number of tiles of each players by the way
*/

short i,j,x,y,k;
unsigned short r,u,v,spr;

#ifndef _fargo
void * plane0 = GrayDBufGetHiddenPlane(LIGHT_PLANE);
void * plane1 = GrayDBufGetHiddenPlane(DARK_PLANE);
#else
void * plane0 = gray4lib__plane0;
void * plane1 = gray4lib__plane1;
#endif

	#ifndef _fargo
  memset(plane0,0,3840);
  memset(plane1,0,3840);		// clear both planes
  #else
  clr_scr(plane0);
  clr_scr(plane1);
  #endif


  total[PLAYER1] = total[PLAYER2] = 0;

  for(j=TABLE_HEIGHT;j--;)
    for(i=TABLE_WIDTH-(r=j&1);i--;)
    {
      x = 7*(X+2*i+r);  // calculate coords for a sprite
      y = 7*(Y+j);

      k = table[i][j];
      if(k&mask[PLAYER1]) total[PLAYER1]++;    // count tiles for player 1
      if(k&mask[PLAYER2]) total[PLAYER2]++;

      if((x<Xmax-14) && (y<Ymax-14) && (x>=0) && (y>=0))  // display only sprites that are visible on the screen
      {
      CustomSprite16x15_or(x,y,(k&COLOR_MASK),plane0,plane1);  // draw a sprite

        k &= (P1_MASK|P2_MASK);
        spr = spriteHG;

				for(v=2;v--;)
					for(u=2;u--;spr--)
						if((((table[i-u+r][j+1-2*v]&(P1_MASK|P2_MASK))!=k)&&k)||(u?!(i+r):i+1==TABLE_WIDTH)||(v?!j:j+1==TABLE_HEIGHT)) CustomSprite16x15_or(x,y,spr,plane0,plane1);

      }
    }
}

static void colorize( short i, short j)
{
	if( i+(j&1)<TABLE_WIDTH && j < TABLE_HEIGHT && i>=0 && j>=0)     // don't go out of the grid :)
	{
	  table[i][j] |= (unsigned char)test_mask;				// marks that tile as tested
		tot--;																					// used only by the AI

	  if(test(i+(j&1),j-1))			colorize(i+(j&1),j-1);    // top right
	  if(test(i-1+(j&1),j-1))		colorize(i-1+(j&1),j-1);  // top left
	  if(test(i+(j&1),j+1))			colorize(i+(j&1),j+1);    // bottom right
	  if(test(i-1+(j&1),j+1))		colorize(i-1+(j&1),j+1);  // bottom left
	}

}

static void color_change(void)
{
/*  |  change to player's current color the color of each tiles of the current player */

unsigned short i,j;

  for(j=TABLE_HEIGHT;j--;)
    for(i=TABLE_WIDTH-(j&1);i--;)
      if(test(i,j)) table[i][j] = (color|mask[player]);
}

void expand(void)
{
/*  |  capture all tiles in contact with the player's zone or inside it
    |  color and player need to be fixed before calling that function
*/

  test = test_color_contact;
  color_change();

  player_color[player] = color;

  test_mask = (mask[player]|TEST_MASK);
  test = test_contact;
  colorize(init_coords[2*player],init_coords[2*player+1]);    // fills the tiles in contact

	tot = 0xFFFF;
  test_mask = FILL_MASK;
  test = test_fill;
  colorize(init_coords[2*((player+1)&1)],init_coords[2*((player+1)&1)+1]);  // capture tiles inside of the zone

  test = test_color_fill;
  color_change();
}

void AI_do(void)
{
/* | save table to a buffer, then simulate the capture of each availabe color from
	 | the current position. In the end, chose one of the colors that captures the greatest
	 | number of tiles.
*/

unsigned short k,n;
unsigned char c,old_color;
unsigned char table_save[22][21];

	old_color = player_color[PLAYER2];	// save current color of player 2 (AI)
	n=color=c=0;

	memcpy(table_save,table,22*21);		// save table to a buffer

	for(k=5;k--;)
	{
		while(color==player_color[PLAYER1] || color==old_color) color++;		// do not test any of current player's colors

		expand();			// simulate a play

	  if((tot>n)||(tot==n&&random(2)))
	  {
	  	n=tot;
	  	c=color;
	  }

		memcpy(table,table_save,22*21);		// restores table for the next test or the real play
 		player_color[PLAYER2]=old_color;	// retores current color
 		color++;													// set next color for the test
	}

	color = c;		// set the color
  expand();			// play
}
